import copy
import logging
from typing import TYPE_CHECKING

from uv_upx.services.dependencies_from_project import get_dependencies_from_project
from uv_upx.services.get_all_pyprojects import get_all_pyprojects_by_project_root_path
from uv_upx.services.handle_groups import handle_py_projects
from uv_upx.services.normalize_paths import get_and_check_path_to_uv_lock
from uv_upx.services.toml import toml_load
from uv_upx.services.updater.check_updating_result import check_updating_result
from uv_upx.services.updater.rollback_updater import rollback_updater
from uv_upx.services.updater.update_lock_file import update_lock_file

if TYPE_CHECKING:
    import pathlib


def run_updater(
    *,
    project_root_path: pathlib.Path,
    #
    dry_run: bool = False,
    verbose: bool = False,
    #
    preserve_original_package_names: bool = False,
    #
    no_sync: bool = False,
) -> None:
    logger = logging.getLogger(__name__)

    uv_lock_path = get_and_check_path_to_uv_lock(project_root_path)
    uv_lock_data = toml_load(uv_lock_path)
    uv_lock_data_copy = copy.deepcopy(uv_lock_data)

    py_projects = get_all_pyprojects_by_project_root_path(project_root_path)
    if verbose:
        logger.info(f"Found {len(py_projects.items)} pyproject.toml files in the workspace.")
        for py_project in py_projects.items:
            logger.info(f"  {py_project.path.as_uri()}")

    py_projects_copy = copy.deepcopy(py_projects)

    is_rollback_needed = dry_run
    rollback_message = "Rolling back to previous state because dry run is enabled."

    try:
        update_lock_file(
            project_root_path,
            no_sync=no_sync,
        )

        if handle_py_projects(
            py_projects=py_projects,
            dependencies_registry=get_dependencies_from_project(workdir=project_root_path),
            #
            dry_run=dry_run,
            verbose=verbose,
            #
            preserve_original_package_names=preserve_original_package_names,
        ):
            logger.info("Updated pyproject.toml files successfully.")

            check_updating_result(
                project_root_path,
                dry_run=dry_run,
                #
                no_sync=no_sync,
            )

        else:
            msg = "No important changes detected. Rolling back to previous state."
            logger.info(msg)
            is_rollback_needed = True
            rollback_message = msg

    except Exception as e:  # noqa: BLE001
        msg = f"Failed to update dependencies: '{e}' Rolling back to previous state."
        logger.error(msg)  # noqa: TRY400
        is_rollback_needed = True
        rollback_message = msg

    try:
        if is_rollback_needed:
            rollback_updater(
                uv_lock_path=uv_lock_path,
                uv_lock_data=uv_lock_data_copy,
                #
                py_projects=py_projects_copy,
                #
                no_sync=no_sync,
            )
            logger.info(rollback_message)
    except Exception as e:  # noqa: BLE001
        msg = f"Failed to rollback: '{e}'"
        logger.error(msg)  # noqa: TRY400



from typing import TYPE_CHECKING

from uv_upx.services.run_uv_related import UvSyncMode, run_uv_lock, run_uv_sync

if TYPE_CHECKING:
    import pathlib


def update_lock_file(
    project_root_path: pathlib.Path,
    *,
    no_sync: bool = False,
) -> None:
    if no_sync:
        # Because we want a fast update. Without triggering build for now.
        run_uv_lock(
            workdir=project_root_path,
            upgrade=True,
        )
    else:
        # Because we want to check build problems also.
        run_uv_sync(
            workdir=project_root_path,
            uv_sync_mode=UvSyncMode.UPGRADE,
        )



from typing import TYPE_CHECKING

from uv_upx.services.run_uv_related import UvSyncMode, run_uv_sync
from uv_upx.services.toml import toml_save

if TYPE_CHECKING:
    import pathlib

    from tomlkit import TOMLDocument

    from uv_upx.services.get_all_pyprojects import PyProjectsRegistry


def rollback_updater(
    *,
    uv_lock_path: pathlib.Path,
    uv_lock_data: TOMLDocument,
    #
    py_projects: PyProjectsRegistry,
    #
    no_sync: bool = False,
) -> None:
    toml_save(uv_lock_path, uv_lock_data)
    for py_project in py_projects.items:
        toml_save(py_project.path, py_project.data)

    if not no_sync:
        run_uv_sync(
            workdir=uv_lock_path.parent,
            uv_sync_mode=UvSyncMode.FROZEN,
        )



from typing import TYPE_CHECKING

from uv_upx.services.dependency_up.update_dependency import update_dependency

if TYPE_CHECKING:
    from uv_upx.services.dependencies_from_project import DependenciesRegistry
    from uv_upx.services.dependency_up.models.changes_list import ChangesList
    from uv_upx.services.dependency_up.models.dependencies_list import TomlBasedDependenciesList


def update_dependencies(
    *,
    deps_sequence_from_config: TomlBasedDependenciesList,
    dependencies_registry: DependenciesRegistry,
    #
    verbose: bool = False,
    #
    preserve_original_package_names: bool = False,
) -> ChangesList:
    """Update the list of dependencies.

    Update it in-place.
    Because we want to preserve the comments.
    """
    changes: ChangesList = []
    for index, dependency in enumerate(deps_sequence_from_config):
        changes_or_none = update_dependency(
            dependencies_registry=dependencies_registry,
            dependency=dependency,
            #
            verbose=verbose,
            #
            preserve_original_package_names=preserve_original_package_names,
        )

        if changes_or_none is not None:
            deps_sequence_from_config[index] = changes_or_none.to_item.get_full_spec()
            changes.append(changes_or_none)

    return changes



import logging
from typing import TYPE_CHECKING

from uv_upx.services.run_uv_related import UvSyncMode, run_uv_lock, run_uv_sync

if TYPE_CHECKING:
    import pathlib


def check_updating_result(
    project_root_path: pathlib.Path,
    *,
    dry_run: bool = False,
    #
    no_sync: bool = False,
) -> None:
    logger = logging.getLogger(__name__)

    if dry_run:
        logger.info("Dry run. No changes were made.")

    elif no_sync:
        run_uv_lock(
            workdir=project_root_path,
        )
    else:
        # Because we want to re-check that all is ok.
        run_uv_sync(
            workdir=project_root_path,
            uv_sync_mode=UvSyncMode.FROZEN,
        )
        logger.info("Synced dependencies successfully.")



from typing import Final

# https://peps.python.org/pep-0440/#version-specifiers

VERSION_OPERATOR_I_GREATER_OR_EQUAL: Final[str] = ">="
VERSION_OPERATORS_I_PUT_IF_DIFFERENT: Final[set[str]] = {VERSION_OPERATOR_I_GREATER_OR_EQUAL}
VERSION_OPERATORS_I_EXPLICIT_IGNORE: Final[set[str]] = {
    # Pinned versions
    "==",
    "===",
    #
    # Upper bounds
    "<",
    "<=",
    #
    # May needed advanced logic here
    "~=",
    #
    # Need to calculate a previous version. Skip for now.
    ">",
    #
    # Special case
    "!=",
}

VERSION_OPERATORS_I_ALL: Final[set[str]] = VERSION_OPERATORS_I_PUT_IF_DIFFERENT | VERSION_OPERATORS_I_EXPLICIT_IGNORE



import concurrent.futures
import logging
from typing import TYPE_CHECKING

from uv_upx.services.collect_dependencies.collect_groups_from_py_project import collect_from_py_project
from uv_upx.services.dependency_up import ChangesList, update_dependencies
from uv_upx.services.toml import toml_save

if TYPE_CHECKING:
    from uv_upx.services.dependencies_from_project import DependenciesRegistry
    from uv_upx.services.get_all_pyprojects import PyProjectsRegistry, PyProjectWrapper


def handle_py_project(
    *,
    dependencies_registry: DependenciesRegistry,
    pyproject_wrapper: PyProjectWrapper,
    #
    dry_run: bool,
    verbose: bool,
    #
    preserve_original_package_names: bool = False,
) -> ChangesList:
    """Handle a single pyproject.toml file."""
    logger = logging.getLogger(__name__)

    data = pyproject_wrapper.data

    changes: ChangesList = []

    for group in collect_from_py_project(data):
        changes_local = update_dependencies(
            deps_sequence_from_config=group.dependencies,
            dependencies_registry=dependencies_registry,
            #
            verbose=verbose,
            #
            preserve_original_package_names=preserve_original_package_names,
        )
        changes.extend(changes_local)

    if changes:
        if dry_run:
            logger.info(f"[Dry Run] Changes detected in {pyproject_wrapper.path.as_uri()}, but not saving.")
        else:
            toml_save(pyproject_wrapper.path, data)
            logger.info(f"Saved changes to {pyproject_wrapper.path.as_uri()}")

            for change in changes:
                logger.info(f"  {change}")

    return changes


def handle_py_projects(
    *,
    dependencies_registry: DependenciesRegistry,
    py_projects: PyProjectsRegistry,
    #
    dry_run: bool,
    verbose: bool,
    #
    preserve_original_package_names: bool = False,
) -> ChangesList:
    """Handle multiple pyproject.toml files."""
    changes: ChangesList = []

    # Note: don't really need threads here in this case.
    #   But this place can be one of the places to use parallelism.

    with concurrent.futures.ThreadPoolExecutor() as worker:
        tasks = [
            worker.submit(
                handle_py_project,
                dependencies_registry=dependencies_registry,
                pyproject_wrapper=py_project,
                #
                dry_run=dry_run,
                verbose=verbose,
                #
                preserve_original_package_names=preserve_original_package_names,
            )
            for py_project in py_projects.items
        ]
        for future in concurrent.futures.as_completed(tasks):
            changes_local = future.result()
            changes.extend(changes_local)

    return changes
